var windowOwnPropertyNames = [
    'Object',
    'Function',
    'Array',
    'Number',
    'parseFloat',
    'parseInt',
    'Infinity',
    'NaN',
    'undefined',
    'Boolean',
    'String',
    'Symbol',
    'Date',
    'Promise',
    'RegExp',
    'Error',
    'EvalError',
    'RangeError',
    'ReferenceError',
    'SyntaxError',
    'TypeError',
    'URIError',
    'JSON',
    'Math',
    'console',
    'Intl',
    'ArrayBuffer',
    'Uint8Array',
    'Int8Array',
    'Uint16Array',
    'Int16Array',
    'Uint32Array',
    'Int32Array',
    'Float32Array',
    'Float64Array',
    'Uint8ClampedArray',
    'BigUint64Array',
    'BigInt64Array',
    'DataView',
    'Map',
    'BigInt',
    'Set',
    'WeakMap',
    'WeakSet',
    'Proxy',
    'Reflect',
    'decodeURI',
    'decodeURIComponent',
    'encodeURI',
    'encodeURIComponent',
    'escape',
    'unescape',
    'eval',
    'isFinite',
    'isNaN',
    'TreeWalker',
    'SVGAElement',
    'Range',
    'NodeList',
    'NodeIterator',
    'NodeFilter',
    'Node',
    'Element',
    'HTMLElement',
    'document',
    'innerheight',
    'innerwidth',
    'outerheight',
    'outerwidth',
    'pageXOffset',
    'pageYOffset',
    'addEventListener',
    'removeEventListener',
    'dispatchEvent',
    'Screen',
    'alert',
    'clearInterval',
    'clearTimeout',
    'valueOf',
    'confirm',
    'scrollTo',
    'setInterval',
    'setTimeout',
    'toString',
    'Float32Array',
    'Proxy',
    'EventTarget',
    'hasOwnProperty',
    'isPrototypeOf',
    'toLocaleString',
    'propertyIsEnumerable',
    '__proto__',
    '__defineGetter__',
    '__lookupGetter__',
    '__defineSetter__',
    '__lookupSetter__'
];

class Sandbox {
    constructor(white = false, properties = ['*'], setting = 'allow-scripts allow-same-origin') {
        this.windowOwnPropertyNames = windowOwnPropertyNames;
        const sandbox = this.sandbox = document.createElement('iframe');
        sandbox.src = 'about:blank';
        sandbox.style.display = 'none';
        this.windowOwnPropertyNames = this.windowOwnPropertyNames.concat(properties);
        this.white = white;
        this.set(setting);
        this.enter();
        this.shadow();
        if (!white)
            this.exit();
        return this;
    }
    static getWindowOwnPropertyNames() {
        return windowOwnPropertyNames;
    }
    static debugger(isee, model = { 'safe-context': 'safe' }) {
        let msg = 'I know the danger!';
        if (isee === msg) {
            this.prototype['open debugger'] = true;
            if (model['safe-context'] === 'unsafe') {
                this.prototype['unsafe-context'] = true;
            }
            else {
                console.warn(`%c Sandbox: Opening the 'debugger' will result in a risk of losing security!`, 'font-size: 32px; color: red');
            }
        }
        else {
            console.log(`Sandbox: Opening 'debugger' input '${msg}'`);
        }
    }
    shadow() {
        const contentWindow = this.sandbox.contentWindow;
        const contentDocument = this.sandbox.contentDocument;
        this.window = contentWindow;
        this.content = this.document = contentDocument;
        if (this.white)
            return this;
        const globalProperty = this.getProtoProperty(this.window);
        const ShadowObject = this.window.Object;
        const allowAllProperty = ~this.windowOwnPropertyNames.indexOf('*');
        this.shadowWindow = new contentWindow.Object();
        for (let k of globalProperty) {
            if (this.window[k] + '' !== '[object Window]' && (allowAllProperty || ~this.windowOwnPropertyNames.indexOf(k))) {
                this.shadowWindow[k] = this.window[k];
            }
            else {
                try {
                    ShadowObject.defineProperty(this.shadowWindow, k, {
                        configurable: true,
                        enumerable: true,
                        writable: true,
                        get() {
                            console.error(`Sandbox Error: Unauthorized use of window['${k}'].`);
                            return null;
                        }
                    });
                    this.delProtoProperty(this.window, k);
                }
                catch (e) {
                    //
                }
            }
        }
        this.shadowWindow.shadowWindow = this.shadowWindow;
        this.shadowWindow.window = this.shadowWindow;
        return this;
    }
    getProtoProperty(object) {
        let propertyNames = Object.getOwnPropertyNames(object);
        if (object['__proto__'])
            return propertyNames.concat(this.getProtoProperty(object['__proto__']));
        return propertyNames;
    }
    delProtoProperty(object, name) {
        if (!object)
            return false;
        if (object['__proto__']) {
            if (object['__proto__'].hasOwnProperty(name)) {
                return delete object['__proto__'][name];
            }
            else {
                return this.delProtoProperty(object['__proto__'], name);
            }
        }
        else {
            return delete object[name];
        }
    }
    set(allow) {
        this.sandbox.setAttribute('sandbox', allow);
    }
    reset(allow) {
        this.exit();
        this.set(allow);
        this.enter();
        this.shadow();
        return this;
    }
    open() {
        this.content.open();
        return this;
    }
    write(context = '<head><meta charset="utf-8"></head>') {
        this.content.write(context);
        return this;
    }
    close() {
        this.content.close();
        return this;
    }
    enter() {
        const documentElement = document.documentElement;
        documentElement.appendChild(this.sandbox);
    }
    exit() {
        if (this['open debugger'])
            return;
        const parentNode = this.sandbox.parentNode;
        parentNode && parentNode.removeChild(this.sandbox);
    }
}

const getObjectType = (object) => {
    const objectStr = Object.prototype.toString.call(object);
    const objectType = (/\s*\[(\w+) (\w+)\]\s*/.exec(objectStr) || []);
    switch (objectType[1]) {
        case 'object':
            return objectType[2];
        case 'native':
            return 'native';
    }
    return 'unknow';
};

var allowProtoProperties = {
    Node: [
        'nodeName',
        'nodeType',
        'textContent'
    ],
    Element: [
        'style',
        'onblur',
        'onfocus',
        'onscroll',
        'offsetWidth',
        'offsetHeight',
        'clientWidth',
        'clientHeight',
        'innerText',
        'setAttribute',
        'removeAttribute',
        'createTextNode',
        'addEventListener',
        'getElementsByTagName'
    ],
    HTMLElement: [],
    HTMLBodyElement: [],
    HTMLDivElement: [],
    HTMLUListElement: [],
    HTMLLIElement: [],
    HTMLVideoElement: [],
    HTMLAudioElement: [],
    HTMLSelectElement: [],
    HTMLOptionElement: [],
    HTMLInputElement: [],
    HTMLSpanElement: [],
    HTMLDListElement: [],
    HTMLFontElement: [],
    HTMLHeadingElement: [],
    HTMLParagraphElement: [],
    HTMLDocument: [
        'documentElement',
        'body'
    ],
    Navigator: [
        'appCodeName',
        'appName',
        'appVersion',
        'language',
        'languages',
        'maxTouchPoints',
        'onLine',
        'platform',
        'product',
        'productSub',
        'userAgent',
        'vendor',
        'vendorSub'
    ],
    UserActivation: [
        'hasBeenActive',
        'isActive'
    ],
    Promise: [
        'then',
        'catch',
        'resolve'
    ],
    Touch: [
        'clientX',
        'clientY',
        'force',
        'identifier',
        'pageX',
        'pageY',
        'radiusX',
        'radiusY',
        'rotationAngle',
        'screenX',
        'screenY',
        'target'
    ],
    TouchEvent: [
        'altKey',
        'bubbles',
        'cancelBubble',
        'cancelable',
        'changedTouches',
        'composed',
        'ctrlKey',
        'currentTarget',
        'defaultPrevented',
        'detail',
        'eventPhase',
        'metaKey',
        'returnValue',
        'shiftKey',
        'sourceCapabilities',
        'targetTouches',
        'timeStamp',
        'touches',
        'type',
        'which'
    ],
    MouseEvent: [
        'altKey',
        'bubbles',
        'button',
        'buttons',
        'cancelBubble',
        'cancelable',
        'clientX',
        'clientY',
        'composed',
        'ctrlKey',
        'defaultPrevented',
        'detail',
        'eventPhase',
        'fromElement',
        'getModifierState',
        'initMouseEvent',
        'layerX',
        'layerY',
        'metaKey',
        'movementX',
        'movementY',
        'offsetX',
        'offsetY',
        'pageX',
        'pageY',
        'relatedTarget',
        'returnValue',
        'screenX',
        'screenY',
        'shiftKey',
        'sourceCapabilities',
        'timeStamp',
        'type',
        'which',
        'x',
        'y'
    ]
};

// ShadowFunction
class ShadowFunction {
    constructor() {
        this.sandbox = new Sandbox(false, this['allow properties'] || ['*']);
        this.shadowWindow = this.sandbox.shadowWindow;
        this.ShadowFunction = this.shadowWindow.Function;
        this.shadowHasOwnProperty = this.shadowWindow.Object.hasOwnProperty;
        this.ShadowObject = this.shadowWindow.Object;
        this.ShadowProxy = this.shadowWindow.Proxy;
        this.allowProtoProperties = allowProtoProperties;
        this.run = () => null;
        this.tracker = (e) => {
            if (typeof (e.name) === 'symbol' || !isNaN(Number(e.name)))
                return 'no';
            switch (e.name) {
                case 'length':
                case '__proto__':
                    break;
                default:
                    return this.log(e);
            }
            return 'no';
        };
        this.freezeShadowProto(this.shadowWindow);
    }
    log(e) {
        return console.log('Event Log:', e);
    }
    static debugger(isee, model = { 'no-Proxy': false }) {
        let msg = 'I know the danger!';
        if (isee === msg) {
            this.prototype['open debugger'] = true;
            if (model['no-Proxy']) {
                this.prototype['open debugger: no-Proxy'] = true;
            }
            if (Sandbox.prototype['unsafe-context'] !== true) {
                console.warn(`%c ShadowFunction: Opening the 'debugger' will result in a risk of losing security!`, 'font-size: 32px; color: red');
            }
        }
        else {
            console.log(`ShadowFunction: Opening 'debugger' input '${msg}'`);
        }
    }
    static assignContextProperties(names = ['*']) {
        this.prototype['allow properties'] = names;
    }
    setAllowProtoProperties(allowProperties) {
        Object.assign(this.allowProtoProperties, allowProperties);
        return this.preShadowFunction.bind(this);
    }
    freezeShadowProto(object) {
        const Object = this.ShadowObject;
        const propNames = Object.getOwnPropertyNames(object);
        // Freeze properties before freezing self
        for (let name of propNames) {
            let value = object[name];
            if (!value)
                break;
            if (!(Object.getOwnPropertyDescriptor(object, name) || {}).writable)
                break;
            typeof value === 'object' && this.freezeShadowProto(value);
        }
        if (!Object.isFrozen(object['prototype']))
            Object.freeze(object['prototype']);
    }
    getAllowProperties(object) {
        return this.getAllowProtoProperties(object).concat(Object.getOwnPropertyNames(object));
    }
    getAllowProtoProperties(object) {
        let propertyNames = this.allowProtoProperties[getObjectType(object)] || [];
        if (object['__proto__'])
            return propertyNames.concat(this.getAllowProtoProperties(object['__proto__']));
        return propertyNames;
    }
    isElementObject(value) {
        if (/HTML(\w+)?Element/.exec(getObjectType(value)))
            return true;
        return false;
    }
    isFixedObject(value) {
        if (this.isElementObject(value))
            return false;
        return typeof value === 'object';
    }
    safeGetter(origin, target, name) {
        if (this['open debugger'] && name === '...') {
            console.warn(`ShadowFunction: Unsafe return in debug mode！`);
            return origin;
        }
        if (this.getAllowProperties(origin).indexOf(name) === -1) {
            this.tracker({
                origin,
                name,
                action: 'read'
            });
            return target[name];
        }
        if (origin[name] && !this.isFixedObject(origin[name]) && origin[name].hasOwnProperty === this.shadowHasOwnProperty) {
            return origin[name];
        }
        return this.proxy(origin[name], origin);
    }
    safeSetter(origin, target, name, value) {
        target[name] = this.proxy(value, target);
        if (this.tracker({
            origin,
            name,
            action: 'write',
            value
        }) === `allow set ${name}`)
            origin[name] = target[name];
        return true;
    }
    puppet(origin) {
        const Proxy = this.ShadowProxy;
        const propNames = Object.getOwnPropertyNames(origin);
        const puppet = new this.ShadowObject();
        propNames.map((name) => {
            if ((Object.getOwnPropertyDescriptor(puppet, name) || { writable: true }).writable)
                puppet[name] = '...';
        });
        return new Proxy(puppet, {
            get: (target, name) => {
                return this.safeGetter(origin, target, name);
            },
            set: (target, name, value) => {
                return this.safeSetter(origin, target, name, value);
            }
        });
    }
    proxy(target, self = null) {
        if (this['open debugger: no-Proxy'] || !target)
            return target;
        switch (typeof (target)) {
            case 'string':
            case 'number':
            case 'boolean':
            case 'undefined':
                return target;
            case 'function':
                return this.safeReturnFunction(target, self);
            case 'object':
                return this.puppet(target);
        }
        return;
    }
    safeReturnFunction(fn, origin) {
        const shadowFunction = new this.ShadowFunction('fn', 'origin', 'proxy', `
      return (function () {
        return function () {
          return proxy(fn.apply(origin, proxy(arguments)))
        }
      })()
    `)(fn, origin, this.proxy.bind(this));
        const propNames = Object.getOwnPropertyNames(fn);
        for (let name of propNames) {
            if ((Object.getOwnPropertyDescriptor(fn, name) || { writable: true }).writable) {
                if (name === 'prototype')
                    continue;
                shadowFunction[name] = this.proxy({ 0: fn[name] })[0];
            }
        }
        return shadowFunction;
    }
    preShadowFunction(scriptStr) {
        this.shadowFunction = new this.ShadowFunction(`
      (function () {
        with (arguments[0].window || {}) {
          with (arguments[0].context || {}) {
            (function () {`
            + scriptStr +
            `}).bind(Object.assign(arguments[0].window || {}, arguments[0].context || {}))()
          }
        }
      })(this)`);
        return this.runShadowFunction.bind(this);
    }
    runShadowFunction(origin) {
        const target = this.proxy(origin);
        const shadowFunction = this.shadowFunction;
        shadowFunction.apply(this.ShadowObject.assign(new this.ShadowObject(), { window: this.shadowWindow, context: target }));
        return {
            run: this.preShadowFunction.bind(this),
            proxy: this.proxy.bind(this),
            sandbox: this.sandbox
        };
    }
}
const shadowFunction = (code, setting, log) => {
    const sfnc = new ShadowFunction();
    if (log)
        sfnc.log = log;
    if (setting) {
        sfnc.run = sfnc.setAllowProtoProperties(setting)(code);
    }
    else {
        switch (typeof (code)) {
            case 'object':
                sfnc.run = sfnc.setAllowProtoProperties(code);
                break;
            case 'string':
                sfnc.run = sfnc.preShadowFunction(code);
                break;
            default:
                throw new Error('ShadowFunction: Uncaught SyntaxError: Unexpected identifier.');
        }
    }
    return sfnc.run;
};
shadowFunction.debugger = ShadowFunction.debugger.bind(ShadowFunction);

const DOCUMENT = document;
let constructorId = 0;
// ShadowDocument
class ShadowDocument {
    constructor(root, template, setting = [{}, {}], log) {
        this.run = () => null;
        this.constructorId = 0;
        this.o = 0;
        this.allowTagName = {
            'DIV': true,
            'B': true,
            'P': true,
            'H1': true,
            'H2': true,
            'H3': true,
            'H4': true,
            'H5': true,
            'DL': true,
            'DT': true,
            'DD': true,
            'EM': true,
            'HR': true,
            'UL': true,
            'LI': true,
            'OL': true,
            'TD': true,
            'TH': true,
            'TR': true,
            'TT': true,
            'NAV': true,
            'SUP': true,
            'SUB': true,
            'SPAN': true,
            'FONT': true,
            'BR': true,
            'STYLE': true,
            'SMALL': true,
            'LABEL': true,
            'TABLE': true,
            'TBODY': true,
            'THEAD': true,
            'TFOOT': true,
            'BUTTON': true,
            'FOOTER': true,
            'HEADER': true,
            'STRONG': true
        };
        this.tracker = (e) => {
            if (typeof (this.log) === 'function') {
                this.log(e);
            }
            else {
                console.log('Event Log:', e);
            }
        };
        this.getRealElement = (element) => {
            return this.TREE[element['uuid']] || new Error(`ShadowFunction: Cannot synchronously read compiled style. Async read after the rendering ends / or use '<HTMLElement>.oncomputed = () => { element.offsetTop }'！`);
        };
        this.TREE = {
            0: root.attachShadow ? root.attachShadow({ mode: 'open' }) : root
        };
        this.SHADOWTREE = {};
        Object.assign(this.allowTagName, setting[0] || setting);
        this.log = log;
        this.constructorId = constructorId++;
        this.shadowFunction = new shadowFunction(setting[1] || {}, undefined, log);
        this.shadowFunction = this.shadowFunction(this.injection())(this.empowerment());
        this.sandbox = this.shadowFunction.sandbox;
        this.shadowWindow = this.sandbox.shadowWindow;
        this.shadowDocumentBody = this.shadowWindow.document.body;
        this.ShadowObject = this.shadowWindow.Object;
        this.ShadowNode = this.shadowWindow.Node;
        this.ShadowElement = this.shadowWindow.Element;
        this.shadowGetAttribute = this.ShadowElement['prototype'].getAttribute;
        this.setShadowObserver();
        this.write(template);
        this.run = this.shadowFunction.run.bind(this);
    }
    empowerment() {
        return {
            getShadowElement: (element) => {
                return this.getRealElement(element);
            },
            getShadowElementProps: (element, propsName) => {
                return this.getRealElement(element)[propsName];
            },
            getComputedStyle: (element, pseudoElt = null) => {
                return getComputedStyle(this.getRealElement(element), pseudoElt);
            },
            setShadowEventListener: (listener, type, callback, opt) => {
                let target = null;
                let eventName = 'on-shadow-' + type + '-' + this.constructorId;
                switch (listener) {
                    case '[object Window]':
                        target = window;
                        break;
                    case '[object HTMLDocument]':
                        target = document;
                        break;
                    case '[object HTMLBodyElement]':
                        target = document.body;
                        break;
                    case '[object HTMLHtmlElement]':
                        target = document.documentElement;
                        break;
                }
                if (!target)
                    return;
                if (!callback)
                    return target.removeEventListener(type, target[eventName]);
                Object.defineProperty(target, eventName, {
                    enumerable: false,
                    configurable: true,
                    writable: false,
                    value: (e) => {
                        callback.apply(callback, [this.createShadowEventObject(e)]);
                    }
                });
                target.addEventListener(type, target[eventName], opt);
            }
        };
    }
    isElementObject(value) {
        if (/HTML(\w+)?Element/.exec(getObjectType(value)))
            return true;
        return false;
    }
    createShadowEventObject(originEvent) {
        const target = {};
        for (let key in originEvent) {
            let value = originEvent[key];
            switch (typeof value) {
                case 'string':
                case 'number':
                case 'boolean':
                case 'undefined':
                    target[key] = value;
                    break;
                case 'function':
                    target[key] = value.bind(originEvent);
                    break;
                case 'object':
                    switch (key) {
                        case 'changedTouches':
                        case 'sourceCapabilities':
                        case 'targetTouches':
                        case 'touches':
                            target[key] = this.createShadowEventObject(value);
                            break;
                        case 'srcElement':
                        case 'target':
                        case 'toElement':
                            if (this.isElementObject(value))
                                target[key] = this.SHADOWTREE[value.uuid];
                            break;
                        default:
                            if (!isNaN(Number(key))) {
                                target[key] = this.createShadowEventObject(value);
                            }
                            else {
                                target[key] = value;
                            }
                            break;
                    }
                    break;
            }
        }
        return target;
    }
    injection() {
        return `
      window.addEventListener = EventTarget.prototype.addEventListener = function (type, callback, options) {
        var target = this
        var targetTypeName = Object.prototype.toString.call(target)
        if (!target['onShadowEventNames']) {
          Object.defineProperty(target, 'onShadowEventNames', {
            enumerable: false,
            configurable: true,
            writable: true,
            value: [type]
          })
        } else {
          target['onShadowEventNames'].push(type)
        }
        if (!target['onshadow' + type]) {
          Object.defineProperty(target, 'onshadow' + type, {
            enumerable: false,
            configurable: true,
            writable: true,
            value: {
              type: type,
              options: options,
              callback: [callback]
            }
          })
          setShadowEventListener(targetTypeName, type, function () {
            var args = arguments
            target['onshadow' + type].callback.map(function (fn) {
              fn.apply(target, args)
            })
          }, options)
        } else {
          target['onshadow' + type].callback.push(callback)
        }
      }
      window.removeEventListener = EventTarget.prototype.removeEventListener = function (type, callback, options) {
        var target = this
        var targetTypeName = Object.prototype.toString.call(target)
        var callbackIndex = this['onshadow' + type].callback.indexOf(callback)
        var eventNameIndex = this['onShadowEventNames'].indexOf(type)
        if (eventNameIndex !== -1) {
          this['onShadowEventNames'].splice(eventNameIndex, 1)
        }
        if (callbackIndex !== -1) {
          this['onshadow' + type].callback.splice(callbackIndex, 1)
        }
        if (this['onshadow' + type].callback.length === 0) {
          setShadowEventListener(targetTypeName, type)
        }
      }
      shadowWindow.Object.defineProperties(shadowWindow.HTMLElement.prototype, {
        'offsetHeight': {
          get () {
            return getShadowElementProps(this, 'offsetHeight')
          }
        },
        'offsetWidth': {
          get () {
            return getShadowElementProps(this, 'offsetWidth')
          }
        },
        'offsetTop': {
          get () {
            return getShadowElementProps(this, 'offsetTop')
          }
        },
        'offsetLeft': {
          get () {
            return getShadowElementProps(this, 'offsetLeft')
          }
        },
        'offsetParent': {
          get () {
            return getShadowElementProps(this, 'offsetParent')
          }
        },
        'ref': {
          get () {
            return getShadowElement(this)
          }
        }
      })
    `;
    }
    setShadowObserver() {
        new MutationObserver((records) => {
            for (let record of records) {
                let target = record.target;
                switch (record.type) {
                    case 'attributes':
                        this.setAttribute(record.attributeName, target);
                        break;
                    case 'characterData':
                        this.setCharacterData(target);
                        break;
                    case 'childList':
                        Array.prototype.forEach.call(record.removedNodes, (node) => {
                            this.walker(this.iterator(node), target, true);
                        });
                        Array.prototype.forEach.call(record.addedNodes, (node) => {
                            this.walker(this.iterator(node), target);
                        });
                        break;
                }
            }
        }).observe(this.shadowDocumentBody, {
            subtree: true,
            attributes: true,
            childList: true,
            characterData: true,
            attributeOldValue: true,
            characterDataOldValue: true
        });
    }
    write(template) {
        this.shadowDocumentBody.innerHTML = `<div>${template}</div>`;
    }
    uuid(node, uuid) {
        uuid = parseInt(node.parentNode ? node.parentNode.uuid || 0 : 0, 10);
        uuid++;
        this.o++;
        uuid = uuid + '.' + this.o;
        if (!node.uuid) {
            this.ShadowObject.defineProperty(node, 'uuid', {
                configurable: false,
                enumerable: false,
                value: uuid
            });
            this.SHADOWTREE[uuid] = node;
        }
        return uuid;
    }
    iterator(nodes) {
        if (nodes.nextNode)
            return nodes;
        return DOCUMENT.createNodeIterator(nodes, NodeFilter.SHOW_ALL, null);
    }
    walker(nodes, target, del = false) {
        let node = nodes.nextNode();
        while (node) {
            this.uuid(node);
            switch (node.nodeType) {
                case Node.ELEMENT_NODE:
                    if (del) {
                        this.removeElement(node, target);
                    }
                    else {
                        this.createElement(node, target);
                    }
                    break;
                case Node.TEXT_NODE:
                    if (del) {
                        this.removeTextNode(node, target);
                    }
                    else {
                        this.createTextNode(node, target);
                    }
                    break;
            }
            node = nodes.nextNode();
            if (!node)
                break;
        }
    }
    getParentId(node, target) {
        return (node.parentNode ? node.parentNode['uuid'] : target['uuid']) || 0;
    }
    createElement(node, target) {
        let uuid = node['uuid'];
        let name = this.ShadowNode['prototype'].cloneNode.call(node).nodeName;
        let puuid = this.getParentId(node, target);
        if (this.TREE[uuid])
            return;
        switch (name) {
            case this.allowTagName[name] ? name : null:
                this.TREE[uuid] = DOCUMENT.createElement(name);
                this.TREE[uuid].uuid = uuid;
                break;
            default:
                this.tracker({
                    tagName: name,
                    action: 'createElement'
                });
                throw new Error(`ShadowFunction: The tag name provided ('${name}') is not a valid name of whitelist.`);
        }
        for (let i = 0; i < node.attributes.length; i++) {
            this.setAttribute(node.attributes[i].name, node);
        }
        node['completedState'] = 'complete';
        this.TREE[puuid].appendChild(this.TREE[uuid]);
        this.createEvent(node);
        this.shadowFunction.run(`
      typeof computed === 'function' && computed(el)
    `)({
            el: this.TREE[uuid],
            computed: node['oncomputed']
        });
    }
    removeElement(node, target) {
        let uuid = node['uuid'];
        let puuid = this.getParentId(node, target);
        if (this.TREE[puuid] && this.TREE[uuid]) {
            this.TREE[puuid].removeChild(this.TREE[uuid]);
        }
        delete this.TREE[uuid];
    }
    createTextNode(node, target) {
        let uuid = node['uuid'];
        let puuid = this.getParentId(node, target);
        let text = node.textContent || '';
        if (this.TREE[uuid])
            return;
        this.TREE[uuid] = DOCUMENT.createTextNode(text);
        this.TREE[uuid].uuid = uuid;
        if (this.TREE[puuid]) {
            this.TREE[puuid].appendChild(this.TREE[uuid]);
        }
    }
    removeTextNode(node, target) {
        let uuid = node['uuid'];
        let puuid = this.getParentId(node, target);
        if (this.TREE[puuid] && this.TREE[uuid]) {
            this.TREE[puuid].removeChild(this.TREE[uuid]);
        }
        delete this.TREE[uuid];
    }
    createEvent(node) {
        let onEvent = node['onShadowEventNames'];
        if (!onEvent)
            return;
        onEvent.map((type) => {
            let shadowEvent = node['onshadow' + type];
            if (shadowEvent) {
                this.TREE[node['uuid']].addEventListener(shadowEvent.type, (event) => {
                    event = this.createShadowEventObject(event);
                    this.shadowFunction.run(`
            for (let i = 0; i < events.length; i++) {
              typeof(events[i]) === 'function' && events[i].apply(node, args)
            }
          `)({ events: shadowEvent.callback.map((fn) => fn.bind(node)), node, args: [event] });
                }, shadowEvent.options);
            }
        });
    }
    setAttribute(name, node) {
        let whiteNode = this.TREE[node['uuid']];
        let shadowNode = this.ShadowNode['prototype'].cloneNode.call(node);
        let tagName = shadowNode.tagName;
        let allow = this.allowTagName[tagName];
        let value = this.shadowGetAttribute.call(shadowNode, name);
        let safeAttr = false;
        if (!whiteNode)
            return;
        switch (name) {
            case 'id':
            case 'name':
            case 'style':
            case 'class':
            case 'width':
            case 'height':
                safeAttr = true;
                break;
            default:
                if (typeof (allow) === 'function' && allow(name, value)) {
                    safeAttr = true;
                }
                else {
                    this.tracker({
                        tagName,
                        attributeName: name,
                        action: 'setAttribute',
                        value
                    });
                    throw new Error(`ShadowFunction: The attribute name provided ('${name} in <${tagName.toLocaleLowerCase()} />') is not a valid name of whitelist.`);
                }
                break;
        }
        if (whiteNode && safeAttr)
            whiteNode.setAttribute(name, value);
    }
    setCharacterData(node) {
        let char = this.TREE[node.uuid];
        if (char)
            char.textContent = node.textContent;
    }
}
const shadowDocument = (root, template, setting = [{}, {}], log) => {
    return new ShadowDocument(root, template, setting, log).run;
};

class BlackDocument {
    constructor(template, root = document.body, csp = `default-src 'self';`, type) {
        this.root = root;
        this.create(template, csp, type);
        // @ts-ignore
        return this.run.bind(this);
    }
    hashSource() {
        return (new Array(8 + Math.round(Math.random() * 8))).join(',').split(',').map(() => {
            return String.fromCharCode(97 + Math.round(Math.random() * 25));
        }).join('');
    }
    injectRequestParentAction(csp) {
        let hash = this.hashSource();
        window.addEventListener('message', (data) => {
            let result = data.data || {};
            let method = result.method;
            let params = result.params;
            let postToSandbox = (data) => {
                if (this.sandbox.contentWindow) {
                    this.sandbox.contentWindow.postMessage({
                        data,
                        key: method + ':' + JSON.stringify(params)
                    }, '*');
                }
            };
            if (result.hash === hash) {
                let func = this.methods[method];
                let promise;
                if (typeof func === 'function') {
                    promise = func(params);
                }
                else {
                    promise = func;
                }
                if (promise instanceof Promise) {
                    promise.then(postToSandbox);
                }
                else {
                    postToSandbox(promise);
                }
            }
        }, false);
        return `
      <meta http-equiv="Content-Security-Policy" content="${csp}">
      <script>
        window.requestParentAction = function (method, params) {
          parent.postMessage({ method: method, params: params, hash: '${hash}' }, '${location.origin}')
          return new Promise(function (reslove, reject) {
            function callback (data) {
              if (data && data.data && data.data.data) {
                if (method + ':' + JSON.stringify(params) === data.data.key) {
                  window.removeEventListener('message', callback)
                  reslove(data.data.data)
                }
              } else {
                reject()
              }
            }
            window.addEventListener('message', callback)
          })
        }
      </script>
    `;
    }
    create(template, csp, type = 'fullscreen') {
        let iframe = document.createElement('iframe');
        iframe.csp = csp;
        iframe.src = URL.createObjectURL(new Blob([`${this.injectRequestParentAction(csp)} ;\n ${template}`], { type: 'text/html' }));
        iframe.setAttribute('sandbox', 'allow-scripts');
        if (type === 'fullscreen')
            iframe.setAttribute('style', `position: fixed; top: 0; right: 0; bottom: 0; left: 0; width: 100%; height: 100%`);
        this.sandbox = iframe;
    }
    run(methods) {
        this.methods = methods;
        this.root.appendChild(this.sandbox);
    }
}

function freeze(object, deep = false, exclude = {}) {
    if (!object)
        return;
    exclude = getObjectType(exclude) === 'Object' ? exclude : {};
    if (deep) {
        let propNames = Object.getOwnPropertyNames(object);
        // Freeze properties before freezing self
        for (let name of propNames) {
            let value = object[name];
            let ignore = exclude[name];
            if (!value)
                break;
            if (isFrozen(value))
                break;
            if (ignore)
                break;
            freeze(value.prototype, true, exclude);
            freeze(value.__proto__, true, exclude);
            let descriptor = Object.getOwnPropertyDescriptor(object, name) || {};
            if (!descriptor.writable)
                break;
            object[name] = value && typeof value === 'object' ?
                freeze(value, true) : value;
        }
    }
    return Object.freeze(object);
}
function isFrozen(obj) {
    return Object.isFrozen(obj);
}

const sandbox = new Sandbox();
const randPassword = function () {
    let shadowString = sandbox.shadowWindow.String;
    let shadowMath = sandbox.shadowWindow.Math;
    let shadowArray = sandbox.shadowWindow.Array;
    return (new shadowArray(8 + shadowMath.round(shadowMath.random() * 8))).join(',').split(',').map(() => {
        return shadowString['fromCharCode'](33 + shadowMath.round(shadowMath.random() * 89));
    }).join('');
};
class Strongbox {
    constructor(insurance, log) {
        this.password = randPassword();
        this.tracker = (e) => {
            if (typeof (this.log) === 'function') {
                this.log(e);
            }
            else {
                console.log('Event Log:', e);
            }
        };
        let that = this;
        let secretKey = this.password;
        // tracker log
        this.log = log;
        for (let insured of insurance) {
            let object = insured[0];
            let props = insured[1];
            let value = object[props];
            Object.defineProperty(object, props, {
                get() {
                    that.password = secretKey = randPassword();
                    return (password, setValue) => {
                        if (password === secretKey) {
                            if (setValue) {
                                value = setValue;
                            }
                            else {
                                return value;
                            }
                        }
                        else {
                            let log = {
                                object,
                                name: props,
                                action: 'read'
                            };
                            if (setValue) {
                                log.value = setValue;
                                log.action = 'write';
                            }
                            that.tracker(log);
                            return;
                        }
                    };
                },
                set() { return; },
                enumerable: true,
                configurable: false
            });
            freeze(object, true);
        }
        return this;
    }
    lock() {
        windowOwnPropertyNames.forEach((key) => {
            let type = typeof (window[key]);
            if (type === 'object' || type === 'function') {
                try {
                    switch (key) {
                        case 'window':
                            break;
                        default:
                            freeze(window[key], true);
                            break;
                    }
                }
                catch (_e) {
                    //
                }
            }
        });
    }
}

class WorkerFunction {
    constructor(code) {
        code = typeof (code) === 'string' ? code : typeof (code) === 'function' ? Object.toString.call(code) : '';
        // code = `self.addEventListener('message', function (event) { ` + code + ` }, false)`
        return new Worker(URL.createObjectURL(new Blob([code], { type: 'text/javascript' })));
    }
}

class TryAgain {
    constructor(task, options) {
        this.task = task;
        this.polls = options.polls || 2;
        this.timeout = options.timeout || 3000;
        this.timeoutId = null;
    }
    try() {
        if (this.polls <= 0)
            return false;
        this.timeoutId = setTimeout(() => {
            if (this.polls-- > 0)
                this.task() && this.try();
        }, this.timeout);
    }
    stop() {
        clearTimeout(this.timeoutId);
    }
    wait(time) {
        this.timeout = time;
    }
    over() {
        this.stop();
        this.polls = 0;
    }
}

const documentElement = document.documentElement;
const clearUnsafe = `
  function clearUnsafe (proto) {
    var props = Object.getOwnPropertyNames(proto)
    props.map(function (key) {
      try {
        proto[key] = null
        delete proto[key]
      } catch (e) {}
    })
  }
  clearUnsafe(Document.prototype)
  clearUnsafe(Function.prototype)
  clearUnsafe(Object.prototype)
  clearUnsafe(HTMLElement.prototype)
  clearUnsafe(Element.prototype)
  clearUnsafe(Node.prototype)
  clearUnsafe(String.prototype)
  clearUnsafe(window)
`;
const getShadowModule = function (url, options = { receiver: '', timeout: 30000, dataType: 'module' }) {
    return new Promise(function (resolve, reject) {
        let { dataType, receiver, timeout } = options;
        let iframe;
        let tryObj;
        let timeoutId;
        timeout = timeout || 30000;
        // 异常尝试
        tryObj = new TryAgain(send, { timeout: 3000, polls: 2 });
        // abort
        function abort() {
            clearTimeout(timeoutId);
            window.removeEventListener('online', send, false);
            iframe.parentElement && documentElement.removeChild(iframe);
        }
        // over
        function over(type) {
            abort();
            tryObj.try();
            if (tryObj.polls === 0) {
                reject(type);
            }
        }
        function hashSource() {
            return (new Array(8 + Math.round(Math.random() * 8))).join(',').split(',').map(() => {
                return String.fromCharCode(97 + Math.round(Math.random() * 25));
            }).join('');
        }
        function send() {
            let hash = hashSource();
            let script = document.createElement('script');
            script.src = url;
            url = script.src;
            iframe = document.createElement('iframe');
            iframe.style.display = 'none';
            iframe.csp = `script-src 'unsafe-inline' ${url.split(/\?|\#/)[0]} 'nonce-${hash}'; ${options.reportUrl ? 'report-uri ' + options.reportUrl : ''}`;
            iframe.src = URL.createObjectURL(new Blob([`
        <html>
          <head>
            <meta http-equiv="Content-Security-Policy" content="${iframe.csp}">
            <script id=${hash} nonce=${hash} src="${url}" async></script>
            <script nonce=${hash}>
              (function () {
                var parent = window.parent
                var script = document.getElementById('${hash}')
                window['${receiver}'] = window['moduleExportsReceiver'] = function (data) {
                  window['moduleExportsReceiver'] = function () {}
                  parent['on${hash}'] ? parent['on${hash}']({ data: { response: data, src: '${url}', hash: '${hash}' } }) : parent.postMessage({ response: JSON.parse(JSON.stringify(data)), src: '${url}', hash: '${hash}' }, '${location.origin}')
                }
                script.onerror = window['getModuleError'] = function () {
                  window['moduleExportsReceiver'](null)
                }
                script.onload = window['getModuleLoaded'] = function () {
                  var data = Object.keys(window.exports).length ? window.exports : window.module.exports
                  window['moduleExportsReceiver'](data)
                }
                ${dataType === 'json' ? clearUnsafe : ''}
              })()
              window.module = {
                exports: {}
              }
              window.exports = {}
            </script>
          <head>
        <body>
        </body>
        </html>
      `], { type: 'text/html' }));
            if (dataType !== 'module') {
                iframe.setAttribute('sandbox', 'allow-scripts');
            }
            window['on' + hash] = function (data) {
                let result = data.data;
                let response = result ? result.response : '';
                // filter message
                if (result.src !== url || result.hash !== hash)
                    return;
                // remove iframe
                abort();
                delete window['on' + hash];
                // resolve | reject
                if (response) {
                    resolve(response);
                }
                else {
                    reject(result);
                }
            };
            window.addEventListener('message', window['on' + hash], false);
            documentElement.appendChild(iframe);
        }
        // 超时处理
        timeoutId = setTimeout(() => {
            over('timeout');
        }, timeout);
        send();
        // 断网重连
        if (navigator.onLine === false) {
            abort();
            tryObj.stop();
            window.addEventListener('online', send, false);
        }
    });
};

const loadShadowScript = (libs, path = '') => {
    let task = [];
    let codeMap = {};
    let code = '';
    for (let url of libs) {
        task.push(new Promise((resolve, reject) => {
            fetch(path + url).then((response) => {
                return response.text();
            }).then((text) => {
                codeMap[url] = text;
                resolve();
            }).catch(() => {
                console.log('load ' + url + ' failed!');
                reject();
            });
        }));
    }
    return new Promise((resolve, reject) => {
        Promise.all(task).then(() => {
            for (let url of libs) {
                code += '; \n' + codeMap[url] + '; \n';
            }
            resolve(code);
        }).catch(() => {
            reject();
        });
    });
};

// snadbox && shadowSandbox
const shadowSandbox = new Sandbox();
const shadowWindow = shadowSandbox.shadowWindow;
const safeEval = shadowWindow.eval;

const object2params = (obj) => {
    let url = '';
    Object.entries(obj).forEach((v) => {
        if (typeof (v[1]) === 'object')
            v[1] = JSON.stringify(v[1]);
        url += v.join('=') + '&';
    });
    return url.substring(0, url.length - 1);
};

const jsonp = function (options) {
    return new Promise(function (resolve, reject) {
        let { url, data, callbackKey, callbackName, timeout, reportUrl } = options;
        let payload = '';
        data = data || {};
        timeout = timeout || 30000;
        callbackKey = callbackKey || 'callback';
        callbackName = callbackName || 'JSONP';
        data[callbackKey] = callbackName;
        payload = object2params(data);
        if (!url || typeof url !== 'string')
            return reject('params url is not defined');
        url += (url.indexOf('?') !== -1 ? '&' : '?') + payload;
        getShadowModule(url, { receiver: callbackName, timeout, reportUrl }).then((data) => {
            resolve(JSON.parse(JSON.stringify(data)));
        }).catch(reject);
    });
};

const csp = (desc) => {
    const meta = document.createElement('meta');
    meta.setAttribute('charset', 'utf-8');
    meta.setAttribute('http-equiv', 'Content-Security-Policy');
    meta.setAttribute('content', desc);
    document.getElementsByTagName('head')[0].appendChild(meta);
};

export { Sandbox, shadowFunction as ShadowFunction, shadowDocument as ShadowDocument, WorkerFunction, BlackDocument, Strongbox, TryAgain, getShadowModule, loadShadowScript, safeEval, jsonp, csp };
